// ------- polygon.h
#ifndef POLYGON_H
#define POLYGON_H
#include <polyline.h>

namespace GraphicsLibrary  {

template <class T>
class Polygon : public Polyline<T>  {
protected:
	bool completed;
	void addlastvertex();
	virtual void drawfilled();
private:
	struct edgestate {
		Point point;
		int wholepixelxmove;
		int xdirection;
		int errorterm;
		int errortermadjup;
		int errortermadjdown;
		int count;
		static bool yxcomp(const edgestate& es1, const edgestate& es2);
		static bool xycomp(const edgestate& es1, const edgestate& es2);
	};
	std::vector<edgestate> edgetable;
	std::vector<edgestate> active_edgetable;
	std::vector<edgestate>::iterator edges;
	void buildedges();
	void yedges_to_active(int ytomove);
	void draw_active_line(int ytodraw);
	void advance_active_edge();
public:
    Polygon(T& t, Point pos, std::vector<Point> verts, Color clr = black, 
					bool fill = false) : 
				Polyline<T>(t, pos, verts, clr), completed(false)
	    { filled = fill; }
    Polygon(T& t, Point pos, Color clr = black, bool fill = false) : 
				Polyline<T>(t, pos, clr), completed(false)
    	{ filled = fill; }
    virtual void draw();
	void rotate(Point& pt);
};

template <class T>
void Polygon<T>::rotate(Point& pt)
{
	edgetable.clear();
	Polyline<T>::rotate(pt);
}

template <class T>
bool Polygon<T>::edgestate::yxcomp(const edgestate& es1, const edgestate& es2)
{
	return es1.point.y == es2.point.y ? es1.point.x < es2.point.x : 
			es1.point.y < es2.point.y;
}
template <class T>
bool Polygon<T>::edgestate::xycomp(const edgestate& es1, const edgestate& es2)
{
	return es1.point.x == es2.point.x ? es1.point.y < es2.point.y : 
			es1.point.x < es2.point.x;
}

// --- add the final, connecting vertex
template <class T>
void Polygon<T>::addlastvertex()
{
	if (vertices.size() > 0)	{
		addvertex(*vertices.begin());
		completed = true;
	}
}

template <class T>
void Polygon<T>::draw()
{
	if (!completed)
		addlastvertex();
	filled ? drawfilled() : Polyline<T>::draw();
}
template <class T>
void Polygon<T>::drawfilled()
{
	if (vertices.size() < 3)	// need at least 3 vertices to draw
		return;
	if (edgetable.size() == 0)
		buildedges();
	if (edgetable.size())	{
		active_edgetable.clear();
		int currenty = edgetable.begin()->point.y;

		edges = edgetable.begin();
		while (edges != edgetable.end() || active_edgetable.size() > 0)	{
			yedges_to_active(currenty);
			draw_active_line(currenty);
			advance_active_edge();
			currenty++;
		}
	}
}
template <class T>
void Polygon<T>::yedges_to_active(int ytomove)
{
	while (edges != edgetable.end() && edges->point.y == ytomove)
		active_edgetable.push_back(*(edges++));
	std::sort(active_edgetable.begin(), active_edgetable.end(), edgestate::xycomp);
}
template <class T>
void Polygon<T>::draw_active_line(int ytodraw)
{
	std::vector<edgestate>::iterator iter;
	for (iter = active_edgetable.begin(); iter < active_edgetable.end(); iter += 2)
		for (int x = iter->point.x; x < (iter+1)->point.x; x++)
			drawpoint(Point(x, ytodraw));
}
// advances each edge in the active edge table by one scan line.
// removes edges that have been fully scanned.
template <class T>
void Polygon<T>::advance_active_edge()
{
	std::vector<edgestate> escopy;
	std::vector<edgestate>::iterator iter;
	for (iter  = active_edgetable.begin(); iter != active_edgetable.end(); iter++)	{
		edgestate es = *iter;
		if (--es.count != 0)	{
			es.point.x += es.wholepixelxmove;
			if ((es.errorterm += es.errortermadjup) > 0) {
				es.point.x += es.xdirection;
				es.errorterm -= es.errortermadjdown;
			}
			escopy.push_back(es);
		}
	}
	active_edgetable = escopy;
}
// ---- build the edge table from the polygon vertices
template <class T>
void Polygon<T>::buildedges()
{
	std::vector<Point>::const_iterator points;
	for (points = vertices.begin(); points != vertices.end(); points++)	{
		// --- calculate edge height and width
		Point begpt(*points);
		Point endpt;
		// --- edge runs from current point to previous one
		if (points == vertices.begin())
			// wrap around to end of the list
			endpt = Point((vertices.end()-1)->x,(vertices.end()-1)->y);
		else
			endpt = Point((points-1)->x, (points-1)->y);
		// --- rotate points
		Polyline<T>::rotate(begpt);
		Polyline<T>::rotate(endpt);
		// --- coerce edge to run top to bottom
		if (begpt.y > endpt.y)
			std::swap(begpt, endpt);
		int deltay = endpt.y - begpt.y;
		if (deltay == 0)
			continue;	// 0 height edge can't be active
		int deltax = endpt.x - begpt.x;
		int width = std::abs(deltax);
		edgestate es;
		es.xdirection = deltax > 0 ? 1 : -1; // direction in which x moves
		es.point = begpt;
		es.count = deltay;
		es.errortermadjdown = deltay;
		if (deltax >= 0)	// initial error term going L->R
            es.errorterm = 0;
         else				// initial error term going R->L
			es.errorterm = -deltay + 1;
		if (deltay >= width)	{
			// --- y-major edge
            es.wholepixelxmove = 0;
            es.errortermadjup = width;
		}
		else	{
			// --- x-major edge
			es.wholepixelxmove = (width / deltay) * es.xdirection;
			es.errortermadjup = width % deltay;
		}
		edgetable.push_back(es);
	}
	// --- sort the edge table by y/x coordinates
	std::sort(edgetable.begin(), edgetable.end(), edgestate::yxcomp);
}

} // namespace GraphicsLibrary

#endif