"""
usage: cg [-t size] [-g] [-l <pythonfile.py>] inputfile

inputfile - file containing the code template
outputfile - file to write the results to
-g - convert inputfile to python without executing.
-l - Python script file to be executed before processing the inputfile.
-t - Number of spaces to use when converting tabs to spaces.
"""

Version = "1.0.1"

import string, sys, os, pickle, tempfile, glob
import UserDict

# Save the standard streams as soon as the module is loaded.
# Hopefully these point to the console since they are used for 
# user input/output functions.
outstream = sys.stdout
instream = sys.stdin

# class Properties
# A UserDict subclass that knows how to read and write its dictionary 
# to a Java like properties file."
class Properties(UserDict.UserDict):

	def __init__(self, defProperties):
		UserDict.UserDict.__init__(self)
		if defProperties != None:
			self.data = defProperties.data
	
	def __getitem__(self, key): 
		return self.data[string.lower(key)]
	def __setitem__(self, key, item): 
		self.data[string.lower(key)] = item
	def __delitem__(self, key): 
		del self.data[string.lower(key)]
	def has_key(self, key): 
		return self.data.has_key(string.lower(key))
	
	def get(self, key, failobj=None):
		if self.data.has_key(string.lower(key)):
			return self.data[string.lower(key)]
		else:
			return failobj
			
	def read( self, fileName ):

		try:
			self.openFileName = fileName
			file = open(fileName, "r")
			self.data = pickle.loads( file.read() )
		except:
			pass
			
	def save(self):
		self.write( self.openFileName )
		
	def write( self, fileName ):

		try:
			file = open(fileName, "w")
			file.write( pickle.dumps( self.data ) )
			file.close()

		except:
			print "WRN: Unable to write property file [" + fileName + "]"

# class InteractiveDict
# A dictionary that prompts the user for a requested value if the key
# does not exist in the dictionary.
class InteractiveDict(Properties):
	
	def __init__(self, defProperties = None):
		Properties.__init__(self, defProperties)

	def getValue(self, key, msg = "", choices = "", default = ""):
		
		if self.has_key(string.lower(key)) == 0:
			# Prompt the user for the key's value
			if len(msg) == 0:
				msg = "Enter value for " + key
				self.data[string.lower(key)] = self.cg.question(msg, choices, default)
			
		return self.data[string.lower(key)]

class cgEngine:

	# These constants are used to keep track of the state
	# of the parser.
	stNormal = 0
	stFoundScriptPrefix1 = 1
	stScriptOpen = 2
	stFoundScriptPostfix1 = 3
	
	def __init__(self, cachedir = '.'):
		
		self.localDict = locals()
		self.props = InteractiveDict()
		self.props.cg = self
		self.state = cgEngine.stNormal

		if cachedir[len(cachedir)-1] == os.sep:
			cachedir = cachedir[:-1]
			
		self.cachedir = cachedir
		
	def indent( self, pos ): return '\t' * (pos)
	
	def input( self, prompt = "" ):
		"Prompt the user for a text string."
		
		oldInStream = sys.stdin
		oldOutStream = sys.stdout
		
		sys.stdin = instream
		sys.stdout = outstream
		
		s = raw_input(prompt)
		
		sys.stdin = oldInStream
		sys.stdout = oldOutStream
		
		return s

	def question( self, msg, choices = None, default = None ):
		"Ask the user a question via stdout."
	
		display = msg
		if choices != None and len(choices) > 0:
			msg = msg + " (" + choices + ")"
	
		value = self.input( msg + " ? " )
		if value == "" and default != None:
			value = default
	
		return value

	def output( self, filename ):
		"Redirects the template output to the chosen file."
		sys.stdout = open( filename, "w" )

	def cleanup( self ):
		"""Removes any temporary files created by this instance or previous instances
		of cgEngine."""

		filename = string.replace(tempfile.gettempdir() + "/*.cgo", '/', os.sep)
		for name in glob.glob(filename):
			os.unlink( name )

	def makeCacheFile( self, file ):

		cacheFile = file
		driveIndex = string.find(file, ":")
		if driveIndex > -1:
			cacheFile = file[driveIndex+1:]
			
		# make directories if necessary
		pathParts = string.split( os.path.split(cacheFile)[0], os.sep )
		testDir = self.cachedir
		for part in pathParts:
			testDir = testDir + os.sep + part
			if not os.path.exists( testDir ):
				os.mkdir( testDir )
		
		if cacheFile[0] == os.sep:
			cacheFile = self.cachedir + cacheFile
		else:
			cacheFile =  self.cachedir + os.sep + cacheFile

		return cacheFile
		
	def generate( self, inputfile, generateOnly = 0 ):
		"Generate an output file from the chosen template."
		
		inFile = open(inputfile, "r")
		outFileName = self.makeCacheFile( inputfile )
		outFile = open( outFileName, "w" )

		# Save the standard output stream in case the script calls our
		# output method and redirects the output to some other place
		stdoutSave = sys.stdout
		
		self.__convertToPython( inFile, outFile )
		
		inFile.close()
		outFile.close()

		if generateOnly:
			
			return outFileName;
			
		else:
			
			# Save our properties
			oldProps = self.props
			
			# Read in the properties file for this template
			# (if it exists)
			root,ext = os.path.splitext(inputfile)
			propertiesFile = root + ".prp"
			
			self.props = InteractiveDict()
			self.props.cg = self
			self.props.read( propertiesFile )
		
			# Make sure the dictionary contains a link to the code generator
			# services.
			self.localDict["cg"] = self
			exec( "def __write__(s): cg.__write__(s)", self.localDict )
			execfile( outFileName, self.localDict )
		
			# Restore our properties
			self.props = oldProps
			
		os.unlink( outFileName )
		sys.stdout = stdoutSave
		
	def parseLine( self, inBuf ):
		"Parses a text string into a list of tokens.\n"
		"Each token is either a block of text or a python expression.\n"
		"The classes Text and Expression are used to hold the text blocks "
		"and python expressions."
		
		tokens = []
		output = Text()
		expression = Expression()
		nPos = 0
						
		while nPos < len(inBuf):
	
			char = inBuf[nPos]
			
			if char == '$':
				
				if self.state == cgEngine.stNormal:
					
					self.state = cgEngine.stFoundScriptPrefix1
				
				elif self.state == cgEngine.stFoundScriptPostfix1:
					
					self.state = cgEngine.stNormal
					tokens.append( expression )
					expression = Expression()
					
				elif self.state == cgEngine.stFoundScriptPrefix1:
					output.append( "$" )
				elif self.state == cgEngine.stScriptOpen:
					expression.append( char )
				else:
					output.append( char )
					
			elif char == '[':
				
				if self.state == cgEngine.stFoundScriptPrefix1:
					
					self.state = cgEngine.stScriptOpen
					expression = Expression()
					
					if output.empty() == 0:
						tokens.append(output)
						output = Text()
						
				elif self.state == cgEngine.stScriptOpen:
					expression.append( char )
				else:
					output.append( char )
					
			elif char == ']':
				
				if self.state == cgEngine.stScriptOpen:
					
					self.state = cgEngine.stFoundScriptPostfix1

				elif self.state == cgEngine.stFoundScriptPostfix1:
					expression.append( "]" )
				else:
					
					output.append(char)
			
			else:
				
				if self.state == cgEngine.stScriptOpen and char != '\n':
					expression.append( char )
				elif self.state == cgEngine.stFoundScriptPrefix1:
					self.state = cgEngine.stNormal
					output.append("$")
					output.append(char)
				elif self.state == cgEngine.stFoundScriptPostfix1:
					self.state = cgEngine.stScriptOpen
					expression.append("]")
					expression.append(char)
				elif char != '\n':
					output.append(char)
					
			nPos = nPos + 1
			
		# if there is some output left
		if output.empty() == 0:
			tokens.append( output )

		# if there is an expression
		if expression.empty() == 0:
			tokens.append( expression )
			
		# if there are no tokens, this must be a blank line
		if len(tokens) == 0 and self.state != cgEngine.stScriptOpen:
			tokens.append( Text() )

		return tokens
		
	# convertToPython
	# Processes the input file, converting all script expressions into
	# python code and other text into print statements
	def __convertToPython( self, inFile, outFile, tabIndent = 0 ):
		
		inBuf = "spam"
	
		self.state = cgEngine.stNormal
		
		while len(inBuf):
			
			inBuf = inFile.readline()
			tokens = self.parseLine( inBuf )
			
			# find out if this is only a script line, or contains some output
			printLine = 0
			for token in tokens:
				if token.__class__ == Text:
					printLine = 1
					break
					
			# Write out this line of output
			outBuff = Text()
			if printLine == 0 and len(tokens):
				
				s = string.strip( str(tokens[0]) )
				if len(s) and s[len(s)-1] == ":" and s[0] != "#":
					tabIndent = tabIndent + 1
				
				elif "end block" == string.lower(s):
					tabIndent = tabIndent - 1
					if tabIndent < 0: 
						tabIndent = 0
					tokens[0] = Expression("#" + self.indent(tabIndent) + s)
			
				elif string.lower(s)[:8] == "include ":
					stmt,name = string.split( s, " ", 1 )
					path,inFileName = os.path.split( inFile.name )
					if len(path) == 0:
						newInFile = open( name, "r" )
					else:
						newInFile = open( path + os.sep + name, "r" )
					self.__convertToPython( newInFile, outFile, tabIndent )
					newInFile.close()
					tokens[0] = Expression("#" + self.indent(tabIndent) + s)
				
			for token in tokens:
				
				if token.__class__ == Text:
					outBuff.append( ' + "' + escapeQuotes(str(token)) + '"' )
					printLine = 1
					
				if token.__class__ == Expression:
					if printLine:
						outBuff.append( ' + str(' + str(token) + ')' )
					else:
						outBuff.append( str(token) )

			if printLine:
				# Write out the statement, trimming off the leading + sign.
				outFile.write( self.indent( tabIndent ) + "__write__( " + str(outBuff)[2:] + " )\n" )
			else:
				outFile.write( str( outBuff ) + "\n" )

	def __write__( self, str ):
		print str

class Expression:
	"Contains a single python expression fragment parsed from a template file."
	
	def __init__(self, expr = ""):
		self.data = expr
		
	def append(self, s):
		self.data = self.data + s
		
	def empty(self):
		return len(self.data) == 0
		
	def __repr__(self):
		return self.data
		
class Text:
	"Contains a single text fragment parsed from a template file."
	
	def __init__(self):
		self.data = ""

	def append(self, s):
		self.data = self.data + s

	def empty(self):
		return len(self.data) == 0

	def __repr__(self):
		return self.data

	
# escapeQuotes
def escapeQuotes( s ):
	return string.replace( s, '"', '\\"' )
	

if __name__ == "__main__":
	
	import os, sys, getopt, tempfile
	
	# Print usage message and exit
	def usage(*args):
		sys.stdout = sys.stderr
		for msg in args: print msg
		print __doc__
		sys.exit(2)
	
	generateOnly = 0
	cg = cgEngine()
	
	# Process the command line
	try:
		opts, args = getopt.getopt(sys.argv[1:], 'l:g')
					
	except getopt.error, msg:
		usage(msg)

	if len(args) != 1: usage("Unrecognized option.")
	for o, a in opts:
		if o == "-l": execfile( a, cg.localDict )
		if o == "-g": generateOnly = 1
		if o == "-t": cg.defaultTabSize = int(a)
		
	file = cg.generate( args[0], generateOnly )
	
	if generateOnly:
		f = open( file, "r" )
		sys.stdout.write( f.read() )
		f.close()
		os.unlink( file )
